﻿
Public Class Form1

    ' ===========================================================================================
    '  FORM EVENTS
    ' ===========================================================================================
    Private Sub Form1_Load(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Load
        Me.Text = AppTitleAndVersion()
        Load_INI()
    End Sub
    Private Sub Form1_FormClosing(ByVal sender As Object, ByVal e As System.Windows.Forms.FormClosingEventArgs) Handles Me.FormClosing
        Save_INI()
    End Sub
    Private Sub Button1_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Button1.Click
        ReadDataXY()
        ComputeCoeffs()
    End Sub
    Private Sub Button2_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Button2.Click
        Interpolate()
    End Sub
    Private Sub Button3_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Button3.Click
        ReadDataXY()
        ComputeCoeffs()
        ShowCurveWithGoogle()
    End Sub
    Private Sub txt_Degree_TextChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles txt_Degree.TextChanged
        LimitTextBoxNumber(txt_Degree, 0, 9)
    End Sub
    Private Sub TextBoxes_KeyDown(ByVal sender As Object, ByVal e As KeyEventArgs) Handles txt_Interpolations.KeyDown, _
                                                                                           txt_Data.KeyDown, _
                                                                                           txt_Coeffs.KeyDown
        If e.KeyCode = Keys.A AndAlso e.Control Then
            CType(sender, TextBox).SelectAll()
            e.SuppressKeyPress = True
        End If
    End Sub


    ' ===========================================================================================
    '  UTILITIES
    ' ===========================================================================================
    Private Function ReadDouble(ByVal txt As TextBox) As Double
        Dim s As String = txt.Text
        s = s.Trim.Replace(",", ".")
        Return Val(s)
    End Function

    Private Function ReadInt(ByVal txt As TextBox) As Int32
        Dim s As String = txt.Text
        s = s.Trim.Replace(",", ".")
        Return CInt(Val(s))
    End Function

    Friend Function ReplaceMultipleSpacesAndTabAndTrim(ByVal s As String) As String
        If s = "" Then Return s
        s = s.Replace(vbTab, " ")
        While s.Contains("  ")
            s = s.Replace("  ", " ")
        End While
        Return s.Trim
    End Function

    Private Sub LimitTextBoxNumber(ByRef tb As TextBox, ByVal min As Double, ByVal max As Double)
        Dim n As Double = Val(tb.Text)
        If n < min Then tb.Text = min.ToString : tb.SelectionStart = 999
        If n > max Then tb.Text = max.ToString : tb.SelectionStart = 999
    End Sub

    Private Sub ShowCurveWithGoogle()
        Dim s As String = "y = "
        s += Coeffs(0).ToString("0.######;-0.######", GCI) + " "
        For i As Int32 = 1 To Coeffs.Length - 1
            s += Coeffs(i).ToString("+0.######;-0.######", GCI) + " * x^" + i.ToString + " "
        Next
        Process.Start("Chrome.exe", """? " + s + """")
    End Sub


    ' ===========================================================================================
    '  READ DATA X/Y 
    ' ===========================================================================================
    Private x() As Double
    Private y() As Double
    Private Sub ReadDataXY()
        ReDim x(-1)
        ReDim y(-1)
        For i As Int32 = 0 To txt_Data.Lines.Length - 1
            Dim s As String = txt_Data.Lines(i)
            s = s.Replace(";", " ")
            s = s.Replace(",", " ")
            s = ReplaceMultipleSpacesAndTabAndTrim(s)
            Dim sa As String() = s.Split(" "c)
            If sa.Length = 2 Then
                ReDim Preserve x(x.Length)
                ReDim Preserve y(y.Length)
                x(x.Length - 1) = Val(sa(0))
                y(y.Length - 1) = Val(sa(1))
            End If
        Next
    End Sub


    ' ===========================================================================================
    '  COMPUTE COEFFS
    ' ===========================================================================================
    Private Coeffs(-1) As Double

    Private Sub ComputeCoeffs()
        Dim degree As Int32 = ReadInt(txt_Degree)
        If degree < 0 Or degree > 9 Then Return
        Coeffs = FindPolynomialLeastSquaresFit(x, y, degree)
        txt_Coeffs.Text = ""
        For i As Int32 = 0 To Coeffs.Length - 1
            txt_Coeffs.Text += Coeffs(i).ToString("+000.000000;-000.000000", GCI) + vbCrLf
        Next
    End Sub

    ' ===========================================================================================
    '  FIND THE POLINOMIAL LEAST SQUARES FIT  
    ' ===========================================================================================
    Public Function FindPolynomialLeastSquaresFit(ByVal X() As Double, ByVal Y() As Double, ByVal degree As Integer) As Double()
        Dim answer(-1) As Double
        If X.Length <> Y.Length Then Return answer
        '
        ' --------------------------------------------------------- Allocate space for (degree + 1) equations 
        '                                                           with (degree + 2) terms each (including the constant term)
        Dim coeffs(degree, degree + 1) As Double
        ' --------------------------------------------------------- Calculate the coefficients for the equations
        For j As Integer = 0 To degree
            ' ----------------------------------------------------- Calculate the coefficients for the jth equation
            ' ----------------------------------------------------- Calculate the constant term for this equation
            coeffs(j, degree + 1) = 0
            For i As Int32 = 0 To X.Length - 1
                coeffs(j, degree + 1) -= Math.Pow(X(i), j) * Y(i)
            Next
            ' ----------------------------------------------------- Calculate the other coefficients
            For a_sub As Integer = 0 To degree
                ' ------------------------------------------------- Calculate the dth coefficient
                coeffs(j, a_sub) = 0
                For i As Int32 = 0 To X.Length - 1
                    coeffs(j, a_sub) -= Math.Pow(X(i), a_sub + j)
                Next
            Next a_sub
        Next j
        ' --------------------------------------------------------- Solve the equations
        answer = GaussianElimination(coeffs)
        ' --------------------------------------------------------- Return the result converted into a List(Of Double)
        Return answer
    End Function

    ' --------------------------------------------------------------------
    ' Perform Gaussian elimination on these coefficients
    ' --------------------------------------------------------------------
    ' Return the array of values that gives the solution
    ' --------------------------------------------------------------------
    Private Function GaussianElimination(ByVal coeffs(,) As Double) As Double()
        Dim max_equation As Integer = coeffs.GetUpperBound(0)
        Dim max_coeff As Integer = coeffs.GetUpperBound(1)
        For i As Integer = 0 To max_equation
            ' ------------------------------------------------- Use equation_coeffs(i, i) 
            '                                                   to eliminate the ith coefficient in all of the other equations
            ' ------------------------------------------------- Find a row with non-zero ith coefficient
            If (coeffs(i, i) = 0) Then
                For j As Integer = i + 1 To max_equation
                    ' See if this one works.
                    If (coeffs(j, i) <> 0) Then
                        ' ------------------------------------- This one works. Swap equations i and j
                        ' This starts at k = i because all
                        ' coefficients to the left are 0.
                        For k As Integer = i To max_coeff
                            Dim temp As Double = coeffs(i, k)
                            coeffs(i, k) = coeffs(j, k)
                            coeffs(j, k) = temp
                        Next k
                        Exit For
                    End If
                Next j
            End If
            ' ------------------------------------------------- Make sure we found an equation with a non-zero ith coefficient
            Dim coeff_i_i As Double = coeffs(i, i)
            If coeff_i_i = 0 Then
                MessageBox.Show("There is no unique solution for these points.", _
                                "Message from Theremino Curve Fitting", _
                                MessageBoxButtons.OK, _
                                MessageBoxIcon.Asterisk)
            End If
            ' ------------------------------------------------- Normalize the ith equation.
            For j As Integer = i To max_coeff
                coeffs(i, j) /= coeff_i_i
            Next j
            ' ------------------------------------------------- Use this equation value to zero out the other equations' ith coefficients
            For j As Integer = 0 To max_equation
                ' Skip the ith equation.
                If (j <> i) Then
                    ' Zero the jth equation's ith coefficient.
                    Dim coef_j_i As Double = coeffs(j, i)
                    For d As Integer = 0 To max_coeff
                        coeffs(j, d) -= coeffs(i, d) * coef_j_i
                    Next d
                End If
            Next j
        Next i
        ' ---------------------------------------------------- At this point, the ith equation contains 2 non-zero entries:
        '                                                       The ith entry which is 1
        '                                                       The last entry coeffs(max_coeff)
        '                                                      This means Ai = equation_coef(max_coeff).
        Dim solution(max_equation) As Double
        For i As Integer = 0 To max_equation
            solution(i) = coeffs(i, max_coeff)
        Next i
        ' ---------------------------------------------------- Return the solution values
        Return solution
    End Function


    ' ===========================================================================================
    '  INTERPOLATE VALUES
    ' ===========================================================================================
    Private Function Polinomial(ByVal X As Double, ByVal coeffs() As Double) As Double
        Dim a As Double '= 346
        Dim b As Double '= -10.6
        Dim c As Double '= -1.64
        Dim d As Double
        Dim e As Double
        Dim f As Double
        Dim g As Double
        Dim h As Double
        Dim i As Double
        Dim l As Double

        If coeffs.Length >= 1 Then a = coeffs(0)
        If coeffs.Length >= 2 Then b = coeffs(1)
        If coeffs.Length >= 3 Then c = coeffs(2)
        If coeffs.Length >= 4 Then d = coeffs(3)
        If coeffs.Length >= 5 Then e = coeffs(4)
        If coeffs.Length >= 6 Then f = coeffs(5)
        If coeffs.Length >= 7 Then g = coeffs(6)
        If coeffs.Length >= 8 Then h = coeffs(7)
        If coeffs.Length >= 9 Then i = coeffs(8)
        If coeffs.Length >= 10 Then l = coeffs(9)

        Return a + _
               b * X + _
               c * X ^ 2 + _
               d * X ^ 3 + _
               e * X ^ 4 + _
               f * X ^ 5 + _
               g * X ^ 6 + _
               h * X ^ 7 + _
               i * X ^ 8 + _
               l * X ^ 9
    End Function

    Private Sub Interpolate()
        txt_Interpolations.Text = ""
        Dim ValFrom As Double = ReadDouble(txt_FROM)
        Dim ValTo As Double = ReadDouble(txt_TO)
        Dim ValStep As Double = ReadDouble(txt_STEP)
        Dim Epsilon As Double = 0.000001 * Math.Sign(ValStep)
        Dim y As Double
        For x As Double = ValFrom To (ValTo + Epsilon) Step ValStep
            y = Polinomial(x, Coeffs)
            txt_Interpolations.Text += x.ToString("0.0", GCI).PadRight(8) + y.ToString("0.0", GCI) + vbCrLf
        Next
        txt_Interpolations.Text += vbCrLf
    End Sub


    ' ===========================================================================================
    '  FIND THE LINEAR FIT  --- UNUSED
    ' ===========================================================================================
    'Function LeastSquaresLine(ByVal X() As Double, ByVal Y() As Double) As Double()
    '    Dim SumX, SumXSquared, SumY, SumXY As Double
    '    For I As Integer = 0 To X.Length - 1
    '        SumX += X(I)
    '        SumXSquared += X(I) * X(I)
    '        SumY += Y(I)
    '        SumXY += X(I) * Y(I)
    '    Next
    '    Dim Slope, Intercept, Determinant As Double
    '    Determinant = X.Length * SumXSquared - SumX * SumX
    '    Slope = (X.Length * SumXY - SumY * SumX) / Determinant
    '    Intercept = (SumY * SumXSquared - SumX * SumXY) / Determinant
    '    Return New Double() {Slope, Intercept}
    'End Function

End Class
